/* Copyright (c) David Drysdale 1995 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <getopt.h>

#ifndef PI
#define PI 3.14159
#endif
#define GRAYSCALE 2
FILE * infile;
double *data;
double *dataimag;
double timestep=1.0;
double firsttimestep=-2.0;
double pageoffset=0.0;
double lasttickprinttime;
double lasttickprint;
double lastprint;
double time=0.0;
double temp;
char   byte;
char   eof=0;
int    ii;
int    norows;
int    which;

double voffset=0.0;  /* current vertical offset */
double voffmax=650.0;  /* maximum voffset before reset */
double vstep=5.0;    /* amount to increase voffset per step */
double pagewidth=400.0;/* width of the page */
double deltax;   /* amount to move right per x gridpoint */
double scaling=4.0;  /* norm scaling factor */
double originx=72.0;  /* initial origin translate */
double originy=72.0;
double tick=4.0;  /* tick widths */
double fontsize=15.0; /* font point size */
double maxnorm=50.0; /* norm scale */
double printtime=1.0; /* when to print time ticks */
double mintime=0.05; /* Minimum time separation between printed lines */
double theta;
int    nopages=-1;
int    sample=1;
int    grid=1;
double doubled=1.0;
int    lr=1;
int    complexin=0;
char   grayscale=0;
long   whichpage=1;
int    reverse=1.0;

void printusage(int err)
{
  fprintf(stderr, "Usage: replot [-abytrns] filename\n");
  fprintf(stderr, "Creates a postscript visualization of the data in filename.\n");
  fprintf(stderr, "Real data is assumed to be a list of numbers on each line, with\n");
  fprintf(stderr, "the timestep value in quotes at the beginning of the line.\n");
  fprintf(stderr, "Complex data (-c option) is assumed to be one complex number per\n");
  fprintf(stderr, "line, in form (real,imag), followed by a line with the timestep\n");
  fprintf(stderr, "value on it. Options are\n");
  fprintf(stderr, "   -a a    x coordinate for bottom left of plot (default %.2f)\n",
          originx);
  fprintf(stderr, "   -b b    y coordinate for bottom left of plot (default %.2f)\n",
          originy);
  fprintf(stderr, "   -y y    amount to vertically move between steps (default %.2f)\n",
          vstep);
  fprintf(stderr, "   -t t    maximum y coordinate for each page (default %.2f)\n",
          voffmax);
  fprintf(stderr, "   -r r    maximum x coordinate for each line (default %.2f)\n",
          pagewidth);
  fprintf(stderr, "   -n n    scaling factor for norm of function (default %.2f)\n",
          scaling);
  fprintf(stderr, "   -f f    font size for labels (default %.2f)\n", fontsize);
  fprintf(stderr, "   -m m    size of norm scale at side (default %.2f)\n", maxnorm);
  fprintf(stderr, "   -p p    how often to print timestep labels (default %.2f)\n",
          printtime);
  fprintf(stderr, "   -s s    line sampling rate (default %d)\n", sample);
  fprintf(stderr, "   -x x    minimum time between plots (default %.2f)\n", mintime);
  fprintf(stderr, "   -k k    gridpoint sampling rate (default %d)\n", grid);
  fprintf(stderr, "   -d d    put two columns onto one page\n");
  fprintf(stderr, "   -w w    specify number of points per line for stdin\n");
  fprintf(stderr, "   -g      produce gray scale output\n");
  fprintf(stderr, "   -i      produce inverse gray scale output\n");
  fprintf(stderr, "   -c      allow complex input (needs -w option)\n");
  fprintf(stderr, "   -e e    how many pages to output\n");
  fprintf(stderr, "   -q      reverse output\n");
  fprintf(stderr, "Units are in printers points (1/72 inch) for coordinates\n");
  exit(err);
}

main(int argc, char * argv[])
{
  int opt;

  /* check the arguments */
  norows = 0;
  while ((opt = getopt (argc, argv, "a:b:cd:e:f:ghik:m:n:p:qr:s:t:w:x:y:")) != EOF)
    switch (opt)
    {
      case 'a':
        originx = atof(optarg);
        break;
      case 'b':
        originy = atof(optarg);
        break;
      case 'c':
        complexin = 1;
        break;
      case 'd':
        doubled = (double) atoi(optarg);
        if (doubled < 1.0) doubled = 1.0;
        break;
      case 'e':
        nopages=atoi(optarg);
        break;
      case 'f':
        fontsize = atof(optarg);
        break;
      case 'g':
        grayscale = 1;
        break;
      case 'h':
        printusage(0);
        break;
      case 'i':
        grayscale = -1;
        break;
      case 'k':
        grid = atoi(optarg);
        if (grid <= 0) grid = 1;
        break;
      case 'm':
        maxnorm = atof(optarg);
        break;
      case 'n':
        scaling = atof(optarg);
        break;
      case 'p':
        printtime = atof(optarg);
        break;
      case 'q':
        reverse=-1;
        break;
      case 'r':
        pagewidth = atof(optarg);
        break;
      case 's':
        sample = atoi(optarg);
        if (sample <= 0) sample = 1;
        break;
      case 't':
        voffmax = atof(optarg);
        break;
      case 'w':
        norows = atoi(optarg);
        break;
      case 'x':
        mintime = atof(optarg);
        break;
      case 'y':
        vstep = atof(optarg);
        break;
      default:
        printusage(2);
    }
  argc -= optind;
  argv += optind;

  if (grayscale)
  {
  /*  if (complexin)
    {
      fprintf(stderr, "Error: cannot have complex input with gray scale output\n");
      exit(1);
    }*/
    scaling /= GRAYSCALE;
  }
  lastprint = -mintime;

  if ((argc != 1) && (norows==0))
  {
    printusage(0);
  }
  if (norows == 0)
  {
    if (complexin)
    {
      fprintf(stderr, "Must specify -w option with -c option\n");
      exit(1);
    }
    if ((infile = fopen(argv[0], "r")) == NULL)
    { 
      fprintf(stderr, "Error: can't open %s\n", argv[0]);
      exit(1);
    }
    /* find out the number of numbers on the first line. assumed to be the */
    /* same for all subsequent lines */
    while (isspace(byte = fgetc(infile))) ; 
    if (byte != '\"')
    {
      fprintf(stderr, "Warning: expecting timestep tag on line 1\n");
    }
    else
    {
      byte = fgetc(infile);
      while ((byte != '\"') && (byte != '\n') && (byte != EOF))
      {
        byte = fgetc(infile);
      }
      if (byte == '\n')
      {
        fprintf(stderr, "Error: line break in tag on line 1\n");
        exit(1);
      }
      else if (byte == EOF)
      {
        fprintf(stderr, "Error: file ended in tag on line 1\n");
        exit(1);
      }
      byte = fgetc(infile);
      while (isspace(byte = fgetc(infile))) ;
    }
    ungetc(byte, infile);
    while ( (byte != EOF) && (byte != '\n'))
    {
      norows++;
      while (!isspace(byte)) byte=fgetc(infile);
      while (isspace(byte) && byte != '\n') byte=fgetc(infile);
    }
    rewind(infile);
  }
  else if (argc == 1)
  {
    /* We have been told the number of data points --  if the user has given */
    /* us the wrong number things will go horribly wrong later, but in the   */
    /* meantime, if we have been given a filename, open it.                  */ 
    if ((infile = fopen(argv[0], "r")) == NULL)
    { 
      fprintf(stderr, "Error: can't open %s\n", argv[0]);
      exit(1);
    }
  }
  else
  {
    infile = stdin;
  }

  /* fprintf(stderr, "Norows is %d\n", norows); */
  if ((data = (double *) malloc( (size_t) sizeof(double)*norows)) == NULL)
  {
    fprintf(stderr, "Error: cannot allocate enough memory for one line\n");
    exit(1);
  }
  if (complexin)
  {
    if ((dataimag = (double *) malloc( (size_t) sizeof(double)*norows)) == NULL)
    {
      fprintf(stderr, "Error: cannot allocate enough memory for one line\n");
      exit(1);
    }
  }

  /* first output PostScript prolog */
  deltax = pagewidth / ((double) norows - 1.0);
  printf("%%!\n%%%%Title: cgl plot\n%%%%Creator: replot\n");
  /* printf("%%%%CreationDate: %s\n", date);
  printf("%%%%For:\n");
  printf("%%%%Pages: 1\n"); */
  printf("%%%%Options: ");
  if (complexin)
    printf("-c ");
  if (grayscale==1)
    printf("-g ");
  else if (grayscale==-1)
    printf("-i ");
  printf(
"-a %f -b %f -y %f -t %f -r %f -n %f -f %f -m %f -p %f -s %d -d %f -w %d -x %f -k %d\n", 
          originx, originy, vstep, voffmax, pagewidth, scaling, fontsize,
          maxnorm, printtime, sample, doubled, norows, mintime, grid ); 

  printf("%%%%BoundingBox: %f %f %f %f\n", 
         (originx-fontsize*4.0)/doubled,
         originy -(fontsize/3.0),
         originx+pagewidth+fontsize*4.0+18,
         originy+voffmax+scaling*10.0); 
  printf("%%%%EndComments\n");
  if (grayscale)
  {
    if (complexin) {
      printf("/cgd { %% xleft xright graylevel phase gd => xright\n");
      printf("/theta exch def /gray exch def /xr exch def /xl exch def\n");
      printf("newpath xl voff moveto xr xl sub 0 rlineto 0 %f rlineto\n", vstep);
      if (grayscale==-1) {
        printf("xl xr sub 0 rlineto closepath theta 1 gray sub 1 sethsbcolor fill xr } def\n");
      } else {
        printf("xl xr sub 0 rlineto closepath theta gray 1 sethsbcolor fill xr } def\n");
      }
    } else {
      printf("/gd { %% xleft xright graylevel gd => xright\n");
      printf("/gray exch def /xr exch def /xl exch def\n");
      printf("newpath xl voff moveto xr xl sub 0 rlineto 0 %f rlineto\n", vstep);
      if (grayscale==1) {
        printf("xl xr sub 0 rlineto closepath 1 gray sub setgray fill xr } def\n");
      } else {
        printf("xl xr sub 0 rlineto closepath gray setgray fill xr } def\n");
      }
    }
  }
  if (complexin)
  {
    printf("/dd {/y exch def /x exch def 1 1 sethsbcolor \n");
    printf("     moveto x y lineto stroke x y} def\n");
    printf("/theta {360 mul n div} def /colii {n div} def\n");
    printf("/colourcircle { gsave translate /r exch def /n exch def /ii 1 def\n");
    printf("  4 setlinewidth\n");
    printf("  /olddd {moveto 1 1 sethsbcolor lineto stroke} def\n");
    printf("  1 1 n { pop r ii 1 sub theta sin mul\n");
    printf("          r ii 1 sub theta cos mul neg\n");
    printf("          ii colii\n");
    printf("          r ii theta sin mul\n");
    printf("          r ii theta cos mul neg\n");
    printf("          olddd ii 1 add /ii exch def} for\n");
    printf("  0 1 0 sethsbcolor\n");
    printf("  1 setlinewidth\n");
    printf("  0 0 moveto\n");
    printf("  0 1.5 r mul lineto stroke\n");
    printf("  0 1.5 r mul moveto ( 0) show grestore} def\n");
  }
  printf("%lf %lf translate\n", originx/doubled, originy);
  printf("/Times-Roman findfont %f scalefont setfont\n", fontsize/doubled);

  /* Do the norm scale at the right hand side */
  if (fontsize != 0.0) {
    if (!grayscale)
    {
      printf("%lf 0 moveto ", (pagewidth+12)/doubled);
      printf("%lf %lf lineto ", (pagewidth+12)/doubled, (scaling*maxnorm) ); 
      printf("%lf 0 moveto ", (pagewidth+12)/doubled);
      printf("%lf 0 rlineto ", tick/doubled);
      printf("0 %lf rmoveto ", (-fontsize / 3.0) );
      printf("(  %.2f) show\n", 0.0);
      printf("%lf %lf moveto ", (pagewidth+12)/doubled, (scaling*maxnorm) );
      printf("%lf 0 rlineto ", tick/doubled);
      printf("0 %lf rmoveto ", (-fontsize / 3.0) );
      printf("(  %.2f) show\n", maxnorm);
      printf("%lf %lf moveto ", (pagewidth+12)/doubled, (scaling*maxnorm*.5) );
      printf("%lf 0 rlineto ", tick/doubled);
      printf("0 %lf rmoveto ", (-fontsize / 3.0)/doubled );
      printf("(  %.2f) show\n", (maxnorm*.5));
      printf("stroke\n");
    } else {
      if (grayscale==1) {
        printf(
"/grayscl { %% dx2 dy x y => -\n"
"gsave translate /dy exch def /dx exch def\n"
"/h dy 50 div def 5 setlinewidth\n"
"0 1 50 { /ii exch def \n"
"         dx neg ii h mul moveto \n"
"         1 ii 50 div sub setgray\n"
"         2 dx mul 0 rlineto stroke} for\n"
"0 setlinewidth 0 setgray dx neg 0 moveto\n"
"2 dx mul 0 rlineto 0 dy rlineto 2 dx mul neg 0 rlineto 0 dy neg rlineto stroke\n"
"grestore } def\n");
      } else {
        printf(
"/grayscl { %% dx2 dy x y => -\n"
"gsave translate /dy exch def /dx exch def\n"
"/h dy 50 div def 5 setlinewidth\n"
"0 1 50 { /ii exch def \n"
"         dx neg ii h mul moveto \n"
"         ii 50 div setgray\n"
"         2 dx mul 0 rlineto stroke} for\n"
"0 setlinewidth 0 setgray dx neg 0 moveto"
"2 dx mul 0 rlineto 0 dy rlineto 2 dx mul neg 0 rlineto 0 dy neg rlineto stroke"
"grestore } def\n");
      }

      printf("%f %lf %lf 0 grayscl ", tick/doubled, 3.0*maxnorm, 
                                      (pagewidth+12)/doubled); 

      printf("%lf %lf moveto ", (pagewidth+12+tick)/doubled,(-fontsize / 3.0));
      printf("(  %.2f) show\n", 0.0);
      printf("%lf %lf moveto ", (pagewidth+12+tick)/doubled, 
                                (3.0*maxnorm)+(-fontsize / 3.0) );
      printf("(  %.2f) show\n", 1.0/(scaling*GRAYSCALE));
      printf("%lf %lf moveto ", (pagewidth+12+tick)/doubled, 
             (3.0*maxnorm*.5)+((-fontsize / 3.0)/doubled ) );
      printf("(  %.2f) show\n", 1.0/(scaling*2.0*GRAYSCALE));
      printf("stroke\n");
    }
  } /* if fontsize is not zero */
  
  /* do the phase colour circle */
  if (complexin && (fontsize!=0.0)) 
  {
    if (grayscale) {
      printf("50 %f %lf %lf colourcircle\n", 20.0/doubled, 
             (pagewidth+30.0)/doubled, (maxnorm*6.0) );
    } else {
      printf("50 %f %lf %lf colourcircle\n", 20.0/doubled, 
             (pagewidth+30.0)/doubled, (scaling*maxnorm*3.0) );
    }
  }
  if (reverse==-1)
  {
    printf("0 %f translate\n", voffmax);
    vstep= -vstep;
  }

  /***********************/
  /****** main loop ******/
  /***********************/
  which = sample - 1;
  while (!eof)
  {
    /******************************/
    /* check for new page command */
    /******************************/
    byte = fgetc(infile);
    if (byte == '!') { 
        /**************************/
        /* end of page/strip code */
        /**************************/
        while (byte != '\n') byte=fgetc(infile);
        lr++;
        if ( (double) lr > doubled)
          lr = 1;
        if (grayscale) printf("0 setgray ");
        printf("0 0 moveto\n");
        printf("0 %lf lineto stroke\n", (voffset-vstep) );
        if (whichpage == 1)
          printf("0 0 moveto %lf 0 lineto stroke\n", pagewidth/doubled);
        if (lr == 1)
        {
          printf("showpage\n");
          printf("%lf %lf translate\n", originx/doubled, originy);
          if (reverse==-1)
          {
            printf("0 %f translate\n", voffmax);
          }
          whichpage++;
          if ((whichpage > nopages) && (nopages != -1))
            return(0);
        }
        else 
        {
          /* first move origin back to bottom left of page */
          printf("%lf %lf translate\n", -originx/doubled, -originy);
          printf("%lf %lf translate\n", 
                 originx/doubled+(pagewidth+64+8.0*fontsize)/doubled, originy);
        }
        pageoffset = 0.0;
        voffset = 0.0;
        firsttimestep = -2.0; 
        lastprint = -mintime;
    } else {
      ungetc(byte,infile);
    }
    /********************************************/
    /* first read in a line from the input file */
    /********************************************/
    if (complexin)
    {
      /**************************/
      /* Complex number version */
      /**************************/
      for (ii=0; ii<norows; ii++)
      {
        eof |= (fscanf(infile, "(%lf, %lf)\n", &(data[ii]), &(dataimag[ii])) != 2);
      }
      eof |=  (fscanf(infile, "%lf\n", &temp) != 1);
      timestep = temp - time;
      time = temp;
      if (firsttimestep == -2.0)
      {
        /* the first data -- set up things */
        lasttickprinttime=(double)((int)(time/printtime)) * printtime;
        lasttickprint = lasttickprinttime-time;
        timestep = 0.0;
        firsttimestep = -1.0;
      }
      else if (firsttimestep == -1.0)
        firsttimestep = timestep;
    }
    else
    {
      /***********************/
      /* Real number version */
      /***********************/
      while (isspace(byte = fgetc(infile))) ;
      if (byte != '\"') 
      {
        time += timestep;
        ungetc(byte, infile);
      }
      else
      {
        eof != (fscanf(infile, "%lf\"", &temp) != 1);
        timestep = temp - time;
        time = temp;
        if (firsttimestep == -2.0)
        {
          /* the first data -- set up things */
          lasttickprinttime=(double)((int)(time/printtime)) * printtime;
          lasttickprint = lasttickprinttime-time;
          timestep = 0.0;
          firsttimestep = -1.0;
        }
        else if (firsttimestep == -1.0)
          firsttimestep = timestep;
      }
      for (ii=0; ii<norows; ii++)
      {
        eof |= (fscanf(infile, "%lf", (data+ii)) != 1);
      }
    } 
    /*********************************/
    /* now write out postscript code */
    /*********************************/
    if (++which == sample) which = 0;
    if ((!eof) && (which == 0) )
    {
      /***********************/
      /* Print out tickmarks */
      /***********************/
      voffset +=  (vstep * (timestep/firsttimestep));
      while (time >= (lasttickprinttime+printtime))
      {
        /* lasttickprinttime=(double)((int)(time/printtime)) * printtime; */
        lasttickprinttime+=printtime;
        lasttickprint+=printtime; 
        temp=vstep*lasttickprint/(firsttimestep*(double) sample);
        temp-=pageoffset;
        if (fontsize != 0.0) {
          if (grayscale) printf("0 setgray ");
          printf("0 %lf moveto\n", temp);
          printf("(%.2f  ) dup stringwidth pop neg ", lasttickprinttime);   
          printf(" %f rmoveto show\n",(-fontsize/3.0));
          printf("%lf %lf moveto 0 %lf lineto stroke\n", 
                 -tick/doubled, temp,temp);
        }
      }
      /**************************/
      /* Complex number version */
      /**************************/
      if (complexin && (time >= (lastprint+mintime)))
      {
        lastprint = time;
        if (grayscale) {
          printf("/voff %f def\n0 ", voffset);
          for (ii=1; ii<norows; ii+=grid)
          {
            theta = (atan2(dataimag[ii], data[ii]) + PI) / (2.0 * PI); 
            printf("%f %f %f cgd\n", (ii*deltax)/doubled, 
                   (data[ii]*data[ii]+dataimag[ii]*dataimag[ii])*scaling,theta);
          }
          /* Always include the last point on the line */
          if (ii != norows-1+grid)
          {
            theta = (atan2(dataimag[norows-1], data[norows-1]) + PI) / (2.0 * PI); 
            printf("%f %f %f cgd\n", ((norows-1)*deltax)/doubled, 
                   (data[norows-1]*data[norows-1]+
                    dataimag[norows-1]*dataimag[norows-1])*scaling,theta);
          }
          printf("pop 0 1 0 sethsbcolor\n"); 
        } else {
          printf("0 %f %f %f %f dd\n",
             (voffset+ ((data[0]*data[0]+dataimag[0]*dataimag[0])*scaling)),
                  theta, (1*deltax)/doubled, 
             (voffset+ ((data[1]*data[1]+dataimag[1]*dataimag[1])*scaling)));

          for (ii=2; ii<norows; ii+=grid)
          {
            /* First alternative takes colour from right hand point of a pair */
            theta = (atan2(dataimag[ii], data[ii]) + PI) / (2.0 * PI); 
            /* Second alternative averages the phase between the two points */
            /*theta = (0.5*(atan2(dataimag[ii], data[ii])
                         +atan2(dataimag[ii-1], data[ii-1])) + PI) / (2.0 * PI);*/
            printf("%.4f %.4f %.4f dd\n",
                    theta, (ii*deltax)/doubled, 
             (voffset+ ((data[ii]*data[ii]+dataimag[ii]*dataimag[ii])*scaling)));
          }
          /* Always include the last point on the line */
          if (ii != norows-1+grid)
          {
            ii = norows-1;
            /* First alternative takes colour from right hand point of a pair */
            theta = (atan2(dataimag[ii], data[ii]) + PI) / (2.0 * PI); 
            /* Second alternative averages the phase between the two points */
            /*theta = (0.5*(atan2(dataimag[ii], data[ii])
                       +atan2(dataimag[ii-1], data[ii-1])) + PI) / (2.0 * PI);*/
            printf("%.4f %.4f %.4f dd\n",
                    theta, (ii*deltax)/doubled, 
             (voffset+ ((data[ii]*data[ii]+dataimag[ii]*dataimag[ii])*scaling)));
          }
          /* remove extra x,y and set color back to black */
          printf("pop pop 0 1 0 sethsbcolor\n"); 
        } /* not grayscale version */
      } /* if complexin and its time to print again */
      /***********************/
      /* Real number version */
      /***********************/
      else if (time >= (lastprint+mintime))
      {
        lastprint = time;
        if (grayscale)
        {
          printf("/voff %f def\n0 ", voffset);
          for (ii=1; ii<norows; ii+=grid)
          {
            printf("%f %f gd\n", (ii*deltax)/doubled, 
                   data[ii]*scaling);
          }
          /* Always include the last point on the line */
          if (ii != norows-1+grid)
          {
            printf("%f %f gd\n", ((norows-1)*deltax)/doubled, 
                   data[norows-1]*scaling);
          }
          printf("pop\n");
        }
        else
        {
          printf("0 %f moveto\n", (voffset+(data[0]*scaling)));
          for (ii=1; ii<norows; ii+=grid)
          {
            printf("%f %f lineto\n", (ii*deltax)/doubled, 
                   (voffset+(data[ii]*scaling)));
          }
          /* Always include the last point on the line */
          if (ii != norows-1+grid)
          {
            printf("%f %f lineto\n", ((norows-1)*deltax)/doubled, 
                   (voffset+(data[norows-1]*scaling)));
          }
          printf("stroke\n");
        }
      } /* not complexin and its time to print again */

      /********************/
      /* End of page code */
      /********************/
      if (fabs(voffset) >= voffmax)
      {
        lr++;
        if ( (double) lr > doubled)
          lr = 1;
        if (grayscale) printf("0 setgray ");
        printf("0 0 moveto\n");
        printf("0 %lf lineto stroke\n", (voffset-vstep) );
        if (whichpage == 1)
          printf("0 0 moveto %lf 0 lineto stroke\n", pagewidth/doubled);
        if (lr == 1)
        {
          printf("showpage\n");
          printf("%lf %lf translate\n", originx/doubled, originy);
          whichpage++;
          if ((whichpage > nopages) && (nopages != -1))
            return(0);
        }
        else 
        {
          /* first move origin back to bottom left of page */
          printf("%lf %lf translate\n", -originx/doubled, -originy);
          printf("%lf %lf translate\n", 
                 originx/doubled+(pagewidth+64+8.0*fontsize)/doubled, originy);
        }
        pageoffset +=voffset;
        voffset =  -(vstep * (timestep/firsttimestep));
        /* voffset = 0.0; */
      } /* if voffset is high enough up */
    } /* if not end of file, and this is a sampled timestep */
  } /* while its not the end of the file */
  /****************************/
  /* output PostScript epilog */
  /****************************/
  if (voffset != 0.0)
  {
    if (grayscale) printf("0 setgray ");
    printf("0 0 moveto\n");
    printf("0 %lf lineto stroke\n", voffset );
    if (whichpage == 1)
      printf("0 0 moveto %lf 0 lineto stroke\n", pagewidth/doubled);
    printf("showpage\n");
  }
  return(0);
}
